/*
  gmap.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

//#define DEBUG

#include <qwidget.h>
#include <qcursor.h>
#include <qpainter.h>
#include <stdio.h>
#include <stdlib.h>
#include "qgmap.h"
#include "gmap.h"
#include "mainwindow.h"

#include <qdialog.h>
#include <qdatetime.h>

#include "pos.h"
#include "inputdialog.h"


GMap::GMap( LayeredBase *parent, const char *name)
	: LayeredWidget(parent, name)
{
	map = NULL;

	/* connect signals */
	connect(parent, SIGNAL(positionChanged(Pos &)),
			this, SLOT(setPosition(Pos &)));
	connect(parent, SIGNAL(reload(bool)),
			this, SLOT(updateCenter(bool)));
	connect(parent, SIGNAL(updateMap()),
			this, SLOT(updateMap()));

	connect(this, SIGNAL(cacheFlush()),
			parent, SIGNAL(cacheFlush()));
}

GMap::~GMap()
{
	if (map)
		delete map;
}

void GMap::updateCenter(bool force)
{
	int widget_top, widget_left;
	int new_ltop, new_lleft;

	center.normalize();

	// ウィジェットの左上部の地図座標
	widget_top = center.getY()-this->height()/2;
	widget_left = center.getX()-this->width()/2;

	// 読み込み画像の左上の地図座標
	new_ltop = widget_top & ~(IMAGE_SIZE-1);
	new_lleft = widget_left & ~(IMAGE_SIZE-1);

	Pos pos(new_lleft, new_ltop, center.getZoom());

	// 強制読み込み時、スクロールしたときロード時
	if(load_pos != pos || force){

		//全画像ロード
		loadMap(pos);
		//内部で load_left, load_top を更新
	}

	// 再描画指示
	repaint();
}

bool GMap::paintEvent(QPaintEvent *e, QPixmap &scr)
{
	Enter();

	int dx, dy;

	dx = (center.getX() - this->width()/2) - load_pos.getX();
	dy = (center.getY() - this->height()/2) - load_pos.getY();

	QRect offset_rect = e->rect();
	offset_rect.moveBy(dx, dy);

	bitBlt(&scr, e->rect().topLeft(), &offscreen, offset_rect);

	return true;
}

/* 全画像の読み込み */
void GMap::loadMap(Pos loadpos)
{
	int x, y;

	if (!map) {
		map = new Map(this->parent());
		connect(map, SIGNAL(reload(bool)),
				this, SLOT(updateCenter(bool)));
	}

	for(y = 0; y < y_num; y++){
		for(x = 0; x < x_num; x++){
			const QPixmap *p = map->getMap(
				loadpos.offset(x * IMAGE_SIZE, y * IMAGE_SIZE));

			bitBlt(&offscreen, x * IMAGE_SIZE, y * IMAGE_SIZE,
				   p, 0, 0, IMAGE_SIZE, IMAGE_SIZE);

			map->putMap(p);
		}
	}
	load_pos = loadpos;
}

void GMap::showEvent()
{
	Enter();

	resize(parentWidget()->size());

	x_num = (this->width()/IMAGE_SIZE)+2;
	y_num = (this->height()/IMAGE_SIZE)+2;

	offscreen.resize(x_num*IMAGE_SIZE, y_num*IMAGE_SIZE);

	updateCenter(TRUE);
}

void GMap::setPosition(Pos &pos)
{
	Enter();
	if(center == pos)
		return;

	center = pos;
	map->setPosition(pos);
	updateCenter(FALSE);
}

Pos GMap::getPosition()
{
	return center;
}

void GMap::updateMap()
{
	Enter();
	updateCenter(true);
}

void GMap::toggleGrid()
{
	map->request("Grid", "Toggle");
	emit cacheFlush();
	updateCenter(true);
}

bool GMap::keyPressEvent( QKeyEvent * e )
{
	switch(e->key()){
	case Key_Z:
		toggleGrid();
		return true;
	}
	return false;
}

void GMap::request(QString const &req, QString const &arg)
{
	if (req != "menu")
		return;

	if (arg.toInt() == EV_TOGGLE_GRID)
		toggleGrid();
}
